@@ -18,18 +18,24 @@ from trexlib.utils.common.date_util import parse_datetime
 from trexmodel import conf, program_conf
 from google.auth._default import default
 from trexmodel.models.datastore.system_models import Tagging
+from trexmodel.models.datastore.model_decorators import model_transactional
 
 logger = logging.getLogger('model')
+#logger = logging.getLogger('debug')
 
-class MerchantMin(BaseNModel, DictModel, FullTextSearchable):
+class BusinessEntity(BaseNModel, DictModel):
+    company_name                = ndb.StringProperty(required=False)
+    business_reg_no             = ndb.StringProperty(required=False)
+    
+class MerchantMin(BusinessEntity, FullTextSearchable):
     
-    company_name                = ndb.StringProperty(required=True)
     contact_name                = ndb.StringProperty(required=False)
     address                     = ndb.StringProperty(required=False)
     office_phone                = ndb.StringProperty(required=False)
     mobile_phone                = ndb.StringProperty(required=False)
     fax_phone                   = ndb.StringProperty(required=False)
     email                       = ndb.StringProperty(required=False)
+    website                     = ndb.StringProperty(required=False)
     country                     = ndb.StringProperty(required=False, default='my')
     status                      = ndb.StringProperty(required=False)
     
@@ -51,7 +57,8 @@ class MerchantAcct(MerchantMin):
     logo_public_url                         = ndb.StringProperty(required=False)
     logo_storage_filename                   = ndb.StringProperty(required=False)
     dashboard_stat_figure                   = ndb.JsonProperty()
-    currency_code                           = ndb.StringProperty(required=False, default='myr')
+    currency_code                           = ndb.StringProperty(required=False, default='MYR')
+    locale                                  = ndb.StringProperty(required=False, default='en_MY')
     api_key                                 = ndb.StringProperty(required=False)
     
     
@@ -64,13 +71,44 @@ class MerchantAcct(MerchantMin):
     
     prepaid_configuration                   = ndb.JsonProperty()
     
+    product_modifier_configuration          = ndb.JsonProperty()
+    
     stat_figure_update_interval_in_minutes  = conf.MERCHANT_STAT_FIGURE_UPDATE_INTERVAL_IN_MINUTES
     stat_figure_update_datetime_format      = '%d-%m-%Y %H:%M:%S'
     
-    dict_properties = ['company_name', 'contact_name', 'mobile_phone', 'office_phone', 'fax_phone', 'email', 'account_code', 'country',
+    dict_properties = ['company_name', 'contact_name', 'business_reg_no', 'mobile_phone', 'office_phone', 'fax_phone', 'email', 'account_code', 'country',
                        'registered_datetime', 'modified_datetime', 'plan_start_date', 'plan_end_date', 'currency_code', 
                        'published_program_configuration', 'published_voucher_configuration', 'membership_configuration', 
-                       'tier_membership_configuration', 'prepaid_configuration']
+                       'tier_membership_configuration', 'prepaid_configuration', 'product_modifier_configuration']
+    
+    
+    def to_login_dict(self):
+        return {
+                'account_code'      : self.account_code,
+                'plan_end_date'     : self.plan_end_date,
+                'currency_code'     : self.currency_code,
+                'country'           : self.country,
+                }
+    @property
+    def account_settings(self):
+        return {
+                'account_code'  : self.account_code,
+                'currency'      : self.currency_code,
+                'locale'        : self.locale,    
+                }
+    
+    
+    @staticmethod
+    def update_details(merchant_acct, company_name=None, business_reg_no=None, contact_name=None, email=None, mobile_phone=None, office_phone=None, currency_code=None, country=None):
+        merchant_acct.company_name      = company_name
+        merchant_acct.business_reg_no   = business_reg_no
+        merchant_acct.contact_name      = contact_name
+        merchant_acct.email             = email
+        merchant_acct.mobile_phone      = mobile_phone
+        merchant_acct.office_phone      = office_phone
+        merchant_acct.currency_code     = currency_code
+        merchant_acct.country           = country
+        merchant_acct.put()
     
     @staticmethod
     def format_account_code(account_code):
@@ -209,26 +247,6 @@ class MerchantAcct(MerchantMin):
         
         self.put()                
         
-    def flush_dirty_voucher_configuration(self):
-        if self.published_voucher_configuration and len(self.published_voucher_configuration)>0:
-            existing_vouchers_list  = self.published_voucher_configuration.get('vouchers')
-            new_vouchers_list       = []
-            for p in existing_vouchers_list:
-                try:
-                    program = ndb.Key(urlsafe=p.get('program_key')).get()
-                    if program.archived is False:
-                        new_vouchers_list.append(p)
-                except:
-                    pass
-            
-            self.published_voucher_configuration['vouchers']    = new_vouchers_list
-            self.published_voucher_configuration['count']       = len(new_vouchers_list)
-            
-        else:
-            self.published_voucher_configuration = {'vouchers':[], 'count':0}
-        
-        self.put()    
-    
     def update_published_program(self, program_configuration):
         if is_empty(self.published_program_configuration):
             self.published_program_configuration = {
@@ -305,7 +323,33 @@ class MerchantAcct(MerchantMin):
             self.prepaid_configuration['programs']        = existing_prepaid_program_list
             self.prepaid_configuration['count']           = len(existing_prepaid_program_list) 
             
-        self.put()     
+        self.put()
+        
+    def update_product_modifier(self, product_modifier_configuration):
+        if self.product_modifier_configuration is None or len(self.product_modifier_configuration)==0:
+            self.product_modifier_configuration = {
+                                                'modifiers'  :[product_modifier_configuration],
+                                                'count'     : 1,
+                                                } 
+                                            
+        else:
+            existing_product_modifiers_list  = self.product_modifier_configuration.get('modifiers')
+            
+            modifier_key = product_modifier_configuration.get('modifier_key')
+            
+            index = 0
+            for v in existing_product_modifiers_list:
+                if v.get('modifier_key') == modifier_key:
+                    existing_product_modifiers_list.pop(index)
+                
+                index = index+1
+            
+            existing_product_modifiers_list.append(product_modifier_configuration)
+            
+            self.product_modifier_configuration['modifiers']        = existing_product_modifiers_list
+            self.product_modifier_configuration['count']            = len(existing_product_modifiers_list) 
+            
+        self.put()         
         
     def add_membership(self, membership_configuration):
         if is_empty(self.membership_configuration):
@@ -418,18 +462,16 @@ class MerchantAcct(MerchantMin):
         self.put() 
         
     def remove_archieve_voucher(self, archieve_voucher_key):
-        self.flush_dirty_voucher_configuration()
         existing_vouchers_list = self.published_voucher_configuration['vouchers']
         
-        index = 0
-        
+        new_vouchers_list = []
         for voucher in existing_vouchers_list:
-            if voucher.get('voucher_key') == archieve_voucher_key:
-                existing_vouchers_list.pop(index)
-            index = index+1
+            if voucher.get('code') != archieve_voucher_key:
+                new_vouchers_list.append(voucher)
+            
         
-        self.published_voucher_configuration['vouchers']    = existing_vouchers_list
-        self.published_voucher_configuration['count']       = len(existing_vouchers_list)
+        self.published_voucher_configuration['vouchers']    = new_vouchers_list
+        self.published_voucher_configuration['count']       = len(new_vouchers_list)
             
         self.put()
         
@@ -498,7 +540,42 @@ class MerchantAcct(MerchantMin):
             self.prepaid_configuration['programs']    = existing_programs_list
             self.prepaid_configuration['count']       = program_count
                 
-            self.put()                  
+            self.put() 
+            
+    def remove_product_modifier_configuration(self, modifier_key_to_remove):
+        
+        logger.debug('remove_product_modifier_configuration: modifier_key_to_remove=%s', modifier_key_to_remove)
+        if self.product_modifier_configuration and self.product_modifier_configuration.get('modifiers'):
+            existing_product_modifiers_list     = self.product_modifier_configuration['modifiers']
+            modifiers_count                     = len(existing_product_modifiers_list)
+            
+            logger.debug('modifiers_count before remove=%s', modifiers_count)
+            
+            index = 0
+            
+            for modifier in existing_product_modifiers_list:
+                
+                logger.debug('modifier_key=%s', modifier.get('modifier_key'))
+                
+                is_same_modifier_key = modifier.get('modifier_key') == modifier_key_to_remove
+                
+                logger.debug('is_same_modifier_key=%s', is_same_modifier_key)
+                
+                if is_same_modifier_key:
+                    existing_product_modifiers_list.pop(index)
+                    
+                    logger.debug('Found modifier to be remove')
+                    
+                index = index+1
+            
+            modifiers_count = len(existing_product_modifiers_list)
+            
+            logger.debug('modifiers_count after remove=%s', modifiers_count)
+            
+            self.product_modifier_configuration['modifiers']     = existing_product_modifiers_list
+            self.product_modifier_configuration['count']         = modifiers_count
+                
+            self.put()                          
     
     def update_stat_details(self, stat_dict):
         dashboard_stat_figure = self.dashboard_stat_figure
@@ -595,26 +672,36 @@ class MerchantSentEmail(SentEmail):
     '''
     pass
 
-class Outlet(BaseNModel, DictModel, FullTextSearchable):
+class Outlet(BusinessEntity, FullTextSearchable):
     '''
     Merchant account as Ancestor
     '''
     
-    merchant_acct           = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct)
-    name                    = ndb.StringProperty(required=True)
-    address                 = ndb.StringProperty(required=False)
-    office_phone            = ndb.StringProperty(required=False)
-    fax_phone               = ndb.StringProperty(required=False)
-    email                   = ndb.StringProperty(required=False)
-    business_hour           = ndb.StringProperty(required=False)
-    is_physical_store       = ndb.BooleanProperty(required=False, default=True)
-    geo_location            = ndb.GeoPtProperty(required=False)
-    created_datetime        = ndb.DateTimeProperty(required=True, auto_now_add=True)
+    merchant_acct                   = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct)
+    id                              = ndb.StringProperty(required=False)
+    name                            = ndb.StringProperty(required=True)
+    address                         = ndb.StringProperty(required=False)
+    office_phone                    = ndb.StringProperty(required=False)
+    fax_phone                       = ndb.StringProperty(required=False)
+    email                           = ndb.StringProperty(required=False)
+    business_hour                   = ndb.StringProperty(required=False)
+    is_physical_store               = ndb.BooleanProperty(required=False, default=True)
+    geo_location                    = ndb.GeoPtProperty(required=False)
+    created_datetime                = ndb.DateTimeProperty(required=True, auto_now_add=True)
+    modified_datetime               = ndb.DateTimeProperty(auto_now=True)
+    
+    assigned_tax_setup              = ndb.JsonProperty()
+    receipt_settings                = ndb.JsonProperty()
+    service_charge_settings         = ndb.JsonProperty()
+    
+    assigned_catalogue_key          = ndb.StringProperty(required=False)
+    assigned_dinning_table_list     = ndb.JsonProperty()
     
     fulltextsearch_field_name   = 'name'
     
-    dict_properties         = ['key', 'name', 'address', 'office_phone', 
-                                'fax_phone', 'email', 'business_hour', 'is_physical_store', 
+    dict_properties         = ['key', 'id', 'company_name', 'business_reg_no', 'name', 'address', 'office_phone', 
+                                'fax_phone', 'email', 'business_hour', 'is_physical_store', 'assigned_catalogue_key', 'assigned_dinning_table_list',
+                                'service_charge_settings', 'assigned_tax_setup',
                                 'geo_location', 'created_datetime']
     
     @property
@@ -625,18 +712,91 @@ class Outlet(BaseNModel, DictModel, FullTextSearchable):
     def merchant_acct_key(self):
         return self.key.parent().urlsafe().decode('utf-8')
     
-    
     @property
     def outlet_key(self):
         return self.key.urlsafe()
     
     @staticmethod
-    def create(merchant_acct=None,name=None, address=None, email=None, fax_phone=None, 
-               office_phone=None, business_hour=None, geo_location=None, is_physical_store=True):
+    def assign_catalogue_to_outlet(outlet_key, catalogue_key):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            outlet.assigned_catalogue_key = catalogue_key
+            outlet.put()
+            
+    @staticmethod
+    def remove_catalogue_from_outlet(outlet_key):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            outlet.assigned_catalogue_key = None
+            outlet.put()
+    
+    @staticmethod
+    def add_tax_setup(outlet_key, tax_setup):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            
+            if outlet.assigned_tax_setup:
+                outlet.assigned_tax_setup[tax_setup.get('tax_reg_id')] = tax_setup
+            else:
+            
+                assigned_tax_setup={
+                                tax_setup.get('tax_reg_id'): tax_setup
+                                }
+                outlet.assigned_tax_setup = assigned_tax_setup
+            
+            outlet.put()
+                
+    @staticmethod
+    def remove_tax_setup(outlet_key, tax_reg_id):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            logger.debug('Going to remove tax setup from %s', outlet.name)
+            
+            if outlet.assigned_tax_setup:
+                del outlet.assigned_tax_setup[tax_reg_id]
+            else:
+                outlet.assigned_tax_setup = {}
+            
+            outlet.put()
+            
+    @staticmethod
+    def set_receipt_settings(outlet_key, receipt_settings):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            outlet.receipt_settings = receipt_settings
+            outlet.put()
+                
+    @staticmethod
+    def remove_receipt_settings(outlet_key):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            outlet.receipt_settings = {}
+            outlet.put()        
+            
+    @staticmethod
+    def assign_dinning_table_setup_to_outlet(outlet_key, table_list):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            outlet.assigned_dinning_table_list = table_list.split(',')
+            outlet.put()
+    
+    @staticmethod
+    def remove_dinning_table_setup_from_outlet(outlet_key):
+        outlet = Outlet.fetch(outlet_key)
+        if outlet:
+            outlet.assigned_dinning_table_list = []
+            outlet.put()                
+    
+    @staticmethod
+    def create(merchant_acct=None,name=None, company_name=None, business_reg_no=None, address=None, email=None, fax_phone=None, 
+               office_phone=None, business_hour=None, geo_location=None, is_physical_store=True, id=None):
         
         outlet   = Outlet(
                             parent              = merchant_acct.create_ndb_key(),
-                            name                = name, 
+                            name                = name,
+                            id                  = id,
+                            company_name        = company_name,  
+                            business_reg_no     = business_reg_no,
                             address             = address,
                             email               = email,
                             fax_phone           = fax_phone,
@@ -650,6 +810,24 @@ class Outlet(BaseNModel, DictModel, FullTextSearchable):
         
         return outlet
     
+    @staticmethod
+    def update(outlet, name=None, company_name=None, business_reg_no=None, address=None, email=None, fax_phone=None, 
+               office_phone=None, business_hour=None, geo_location=None, is_physical_store=True, id=None):
+        
+        
+        outlet.name                 = name
+        outlet.id                   = id
+        outlet.company_name         = company_name
+        outlet.business_reg_no      = business_reg_no
+        outlet.email                = email
+        outlet.address              = address
+        outlet.office_phone         = office_phone
+        outlet.fax_phone            = fax_phone
+        outlet.business_hour        = business_hour
+        outlet.geo_location         = geo_location
+        outlet.is_physical_store    = is_physical_store
+        outlet.put()
+    
     @staticmethod
     def list_by_merchant_acct(merchant_acct):
         return Outlet.query(ancestor = merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)
@@ -705,6 +883,22 @@ class MerchantUser(UserMin, FullTextSearchable):
     
     fulltextsearch_field_name   = 'name'
     
+    def to_login_dict(self):
+        return {
+                'user_id'           : self.user_id,
+                'username'          : self.username,
+                'name'              : self.name,
+                'permission'        : self.permission,
+                'granted_outlet'    : self.granted_outlet,
+                'granted_access'    : self.granted_access,
+                'active'            : self.active,
+                'is_admin'          : self.is_admin,
+                'is_admin_user'     : self.is_admin_user,
+                'is_merchant_user'  : self.is_merchant_user,
+                'merchant_acct_key' : self.merchant_acct_key,
+                
+                }
+    
     @property
     def is_super_user(self):
         return False
@@ -779,7 +973,7 @@ class MerchantUser(UserMin, FullTextSearchable):
             return self.permission.get('granted_access')
         else:
             return []
-    
+        
     @staticmethod
     def update_permission(merchant_user, access_permission, outlet_permission, is_admin=False):
         if access_permission is None:
@@ -890,4 +1084,112 @@ class MerchantTagging(Tagging):
     def get_by_merchant_label(merchant_acct, label):
         return MerchantTagging.get_by_label(merchant_acct.create_ndb_key(), label)
     
-    
\ No newline at end of file
+
+class ReceiptSetup(BaseNModel,DictModel):
+    receipt_header_settings = ndb.JsonProperty()
+    receipt_footer_settings = ndb.JsonProperty()
+    modified_datetime       = ndb.DateTimeProperty(auto_now=True)
+    
+    
+    dict_properties = ['receipt_header_settings', 'receipt_footer_settings']
+    
+    @staticmethod
+    def update(merchant_acct, receipt_header_settings={}, receipt_footer_settings={}):
+        receipt_setup = ReceiptSetup.get_by_merchant_acct(merchant_acct)
+        
+        if receipt_setup:
+        
+            receipt_setup.receipt_header_settings      = receipt_header_settings
+            receipt_setup.receipt_footer_settings      = receipt_footer_settings
+            
+        else:
+            receipt_setup = ReceiptSetup(
+                                        parent                  = merchant_acct.create_ndb_key(),
+                                        receipt_header_settings = receipt_header_settings,    
+                                        receipt_footer_settings = receipt_footer_settings,
+                                        )
+        receipt_setup.put() 
+        
+        return receipt_setup   
+    
+    @staticmethod    
+    def remove(receipt_setup):
+        receipt_setup.delete()
+        
+    @staticmethod
+    def get_by_merchant_acct(merchant_acct):
+        return ReceiptSetup.query(ancestor = merchant_acct.create_ndb_key()).get()   
+    
+class ServiceTaxSetup(BaseNModel,DictModel):
+    tax_reg_id              = ndb.StringProperty(required=True)
+    tax_name                = ndb.StringProperty(required=True)
+    tax_label               = ndb.StringProperty(required=True)
+    tax_apply_type          = ndb.StringProperty(required=True)
+    tax_pct_amount          = ndb.FloatProperty(required=True)
+    assigned_outlet_list    = ndb.JsonProperty()
+    is_publish              = ndb.BooleanProperty(default=False)
+    modified_datetime       = ndb.DateTimeProperty(auto_now=True)
+    
+    
+    dict_properties = ['tax_reg_id', 'tax_name', 'tax_label', 'tax_apply_type', 'tax_pct_amount', 'assigned_outlet_list', 'is_publish']
+    
+    @staticmethod
+    def create(tax_reg_id, tax_name, tax_label, tax_apply_type, tax_pct_amount, assigned_outlet_key_list, merchant_acct):
+        service_tax_setup = ServiceTaxSetup(
+                                        parent                  = merchant_acct.create_ndb_key(),
+                                        tax_reg_id              = tax_reg_id,    
+                                        tax_name                = tax_name,
+                                        tax_label               = tax_label,  
+                                        tax_apply_type          = tax_apply_type,    
+                                        tax_pct_amount          = tax_pct_amount,
+                                        is_publish              = False,
+                                        assigned_outlet_list    = assigned_outlet_key_list,
+                                        )
+        service_tax_setup.put()
+        return service_tax_setup
+    
+    @staticmethod
+    def update(service_tax_setup, tax_reg_id, tax_name, tax_label, tax_apply_type, tax_pct_amount, assigned_outlet_key_list):
+        service_tax_setup.tax_reg_id            = tax_reg_id
+        service_tax_setup.tax_name              = tax_name
+        service_tax_setup.tax_label             = tax_label
+        service_tax_setup.tax_apply_type        = tax_apply_type
+        service_tax_setup.tax_pct_amount        = tax_pct_amount
+        service_tax_setup.assigned_outlet_list  = assigned_outlet_key_list
+        service_tax_setup.is_publish            = False
+        service_tax_setup.put()
+    
+    @staticmethod    
+    def remove(service_tax_setup):
+        service_tax_setup.delete()
+        
+    @model_transactional(desc='ServiceTaxSetup.publish')   
+    def publish(self):
+        self.is_publish          = True  
+          
+        for outlet_key in self.assigned_outlet_list:
+            Outlet.add_tax_setup(outlet_key,
+                                    {
+                                     'tax_reg_id'       : self.tax_reg_id,
+                                     'tax_label'        : self.tax_label,
+                                     'tax_apply_type'   : self.tax_apply_type,
+                                     'tax_pct_amount'   : self.tax_pct_amount,   
+                                    })
+            
+        
+        self.put()
+            
+    @model_transactional(desc='ServiceTaxSetup.unpublish')   
+    def unpublish(self):
+        self.is_publish          = False  
+          
+        for outlet_key in self.assigned_outlet_list:
+            Outlet.remove_tax_setup(outlet_key, self.tax_reg_id)
+            
+        
+        self.put()        
+        
+    @staticmethod
+    def list_by_merchant_acct(merchant_acct):
+        return ServiceTaxSetup.query(ancestor = merchant_acct.create_ndb_key()).fetch(limit=conf.MAX_FETCH_RECORD)
+    
